/******************************************************************************
  This file contains routines that can be bound to a Postgres backend and
  called by the backend in the process of processing queries.  The calling
  format for these routines is dictated by Postgres architecture.
******************************************************************************/

#include "postgres.h"

#include <math.h>

#include "access/gist.h"
#include "access/rtree.h"
#include "lib/stringinfo.h"
#include "utils/builtins.h"
#include "utils/varbit.h"

#define abs(a)			((a) <	(0) ? (-a) : (a))

/*
** GiST support methods
*/
bool		g_bit_consistent(GISTENTRY *entry, VarBit * query, StrategyNumber strategy);
GISTENTRY  *g_bit_compress(GISTENTRY *entry);
GISTENTRY  *g_bit_decompress(GISTENTRY *entry);
float	   *g_bit_penalty(GISTENTRY *origentry, GISTENTRY *newentry, float *result);
GIST_SPLITVEC *g_bit_picksplit(GistEntryVector *entryvec, GIST_SPLITVEC *v);
bool		g_bit_leaf_consistent(VarBit * key, VarBit * query, StrategyNumber strategy);
bool		g_bit_internal_consistent(VarBit * key, VarBit * query, StrategyNumber strategy);
VarBit	   *g_bit_union(GistEntryVector *entryvec, int *sizep);
bool	   *g_bit_same(VarBit * b1, VarBit * b2, bool *result);


/*
** R-tree support functions
*/
bool		bit_contains(VarBit * arg1, VarBit * arg2);
bool		bit_contained(VarBit * a, VarBit * b);
bool        bit_eq(VarBit * arg1, VarBit * arg2);
bool		bit_overlap(VarBit * a, VarBit * b);
VarBit	   *bit_union(VarBit * arg1, VarBit * arg2);
VarBit	   *bit_inter(VarBit * a, VarBit * b);
double	   *bit_size(VarBit * a);
void		rt_bit_size(VarBit * a, double *sz);

/*
** These make no sense for this type, but R-tree wants them
*/
bool		bit_over_left(VarBit * a, VarBit * b);
bool		bit_over_right(VarBit * a, VarBit * b);
bool		bit_left(VarBit * a, VarBit * b);
bool		bit_right(VarBit * a, VarBit * b);

/*
** miscellaneous
*/
double	   *bit_distance(VarBit * a, VarBit * b);
int         bit_count_bits(VarBit *arg);
VarBit *    bit_xor(VarBit * arg1, VarBit * arg2);

/*****************************************************************************
 *						   GiST functions
 *****************************************************************************/

/*
** The GiST Consistent method for boxes
** Should return false if for all data items x below entry,
** the predicate x op query == FALSE, where op is the oper
** corresponding to strategy in the pg_amop table.
*/
bool
g_bit_consistent(GISTENTRY *entry,
				  VarBit * query,
				  StrategyNumber strategy)
{
	/*
	 * if entry is not leaf, use g_bit_internal_consistent, else use
	 * g_bit_leaf_consistent
	 */
	if (GIST_LEAF(entry))
		return g_bit_leaf_consistent((VarBit *) DatumGetPointer(entry->key),
									  query, strategy);
	else
		return g_bit_internal_consistent((VarBit *) DatumGetPointer(entry->key),
										  query, strategy);
}


/*
** The GiST Union method for boxes
** returns the minimal bounding box that encloses all the entries in entryvec
*/
VarBit *
g_bit_union(GistEntryVector *entryvec, int *sizep)
{
	int			i;
	VarBit	   *out = (VarBit *) NULL;
	VarBit	   *tmp;

	tmp = (VarBit *) DatumGetPointer(entryvec->vector[0].key);
	for (i = 1; i < entryvec->n; i++) {
		out = bit_union(tmp, (VarBit *) DatumGetPointer(entryvec->vector[i].key));
		if (i > 1)
			pfree(tmp);
		tmp = out;
	}

	*sizep = VARBITBYTES(out);

	return (out);
}

/*
** GiST Compress and Decompress methods for boxes
** do not do anything.
*/
GISTENTRY *
g_bit_compress(GISTENTRY *entry)
{
	return (entry);
}

GISTENTRY *
g_bit_decompress(GISTENTRY *entry)
{
	return (entry);
}



/*
** The GiST Penalty method for boxes
** As in the R-tree paper, we use change in area as our penalty metric
*/
float *
g_bit_penalty(GISTENTRY *origentry, GISTENTRY *newentry, float *result)
{
	VarBit	   *ud;
	double		tmp1,
				tmp2;

	ud = bit_union((VarBit *) DatumGetPointer(origentry->key),
					(VarBit *) DatumGetPointer(newentry->key));
	rt_bit_size(ud, &tmp1);
	rt_bit_size((VarBit *) DatumGetPointer(origentry->key), &tmp2);
	*result = (float) (tmp1 - tmp2);
	pfree(ud);

	return (result);
}



/*
** The GiST PickSplit method for boxes
** We use Guttman's poly time split algorithm
**
** This taken from cube.c which in turn seems to have taken it from other gist
** implementations. I didn't have to do anything more than s/cube/bit/g. Can't
** this be generalized so not every data type needs to provide it?
** 
*/
GIST_SPLITVEC *
g_bit_picksplit(GistEntryVector *entryvec,
				 GIST_SPLITVEC *v)
{
	OffsetNumber i,
				j;
	VarBit	   *datum_alpha,
			   *datum_beta;
	VarBit	   *datum_l,
			   *datum_r;
	VarBit	   *union_d,
			   *union_dl,
			   *union_dr;
	VarBit	   *inter_d;
	bool		firsttime;
	double		size_alpha,
				size_beta,
				size_union,
				size_inter;
	double		size_waste,
				waste;
	double		size_l,
				size_r;
	int			nbytes;
	OffsetNumber seed_1 = 0,
				seed_2 = 0;
	OffsetNumber *left,
			   *right;
	OffsetNumber maxoff;

	/*
	 * fprintf(stderr, "picksplit\n");
	 */
	maxoff = entryvec->n - 2;
	nbytes = (maxoff + 2) * sizeof(OffsetNumber);
	v->spl_left = (OffsetNumber *) palloc(nbytes);
	v->spl_right = (OffsetNumber *) palloc(nbytes);

	firsttime = true;
	waste = 0.0;

	for (i = FirstOffsetNumber; i < maxoff; i = OffsetNumberNext(i))
	{
		datum_alpha = (VarBit *) DatumGetPointer(entryvec->vector[i].key);
		for (j = OffsetNumberNext(i); j <= maxoff; j = OffsetNumberNext(j))
		{
			datum_beta = (VarBit *) DatumGetPointer(entryvec->vector[j].key);

			/* compute the wasted space by unioning these guys */
			/* size_waste = size_union - size_inter; */
			union_d = bit_union(datum_alpha, datum_beta);
			rt_bit_size(union_d, &size_union);
			inter_d = bit_inter(datum_alpha, datum_beta);
			rt_bit_size(inter_d, &size_inter);
			size_waste = size_union - size_inter;

			pfree(union_d);

			if (inter_d != (VarBit *) NULL)
				pfree(inter_d);

			/*
			 * are these a more promising split than what we've already
			 * seen?
			 */

			if (size_waste > waste || firsttime)
			{
				waste = size_waste;
				seed_1 = i;
				seed_2 = j;
				firsttime = false;
			}
		}
	}

	left = v->spl_left;
	v->spl_nleft = 0;
	right = v->spl_right;
	v->spl_nright = 0;

	datum_alpha = (VarBit *) DatumGetPointer(entryvec->vector[seed_1].key);
	datum_l = bit_union(datum_alpha, datum_alpha);
	rt_bit_size(datum_l, &size_l);
	datum_beta = (VarBit *) DatumGetPointer(entryvec->vector[seed_2].key);
	datum_r = bit_union(datum_beta, datum_beta);
	rt_bit_size(datum_r, &size_r);

	/*
	 * Now split up the regions between the two seeds.	An important
	 * property of this split algorithm is that the split vector v has the
	 * indices of items to be split in order in its left and right
	 * vectors.  We exploit this property by doing a merge in the code
	 * that actually splits the page.
	 *
	 * For efficiency, we also place the new index tuple in this loop. This
	 * is handled at the very end, when we have placed all the existing
	 * tuples and i == maxoff + 1.
	 */

	maxoff = OffsetNumberNext(maxoff);
	for (i = FirstOffsetNumber; i <= maxoff; i = OffsetNumberNext(i))
	{
		/*
		 * If we've already decided where to place this item, just put it
		 * on the right list.  Otherwise, we need to figure out which page
		 * needs the least enlargement in order to store the item.
		 */

		if (i == seed_1)
		{
			*left++ = i;
			v->spl_nleft++;
			continue;
		}
		else if (i == seed_2)
		{
			*right++ = i;
			v->spl_nright++;
			continue;
		}

		/* okay, which page needs least enlargement? */
		datum_alpha = (VarBit *) DatumGetPointer(entryvec->vector[i].key);
		union_dl = bit_union(datum_l, datum_alpha);
		union_dr = bit_union(datum_r, datum_alpha);
		rt_bit_size(union_dl, &size_alpha);
		rt_bit_size(union_dr, &size_beta);

		/* pick which page to add it to */
		if (size_alpha - size_l < size_beta - size_r)
		{
			pfree(datum_l);
			pfree(union_dr);
			datum_l = union_dl;
			size_l = size_alpha;
			*left++ = i;
			v->spl_nleft++;
		}
		else
		{
			pfree(datum_r);
			pfree(union_dl);
			datum_r = union_dr;
			size_r = size_alpha;
			*right++ = i;
			v->spl_nright++;
		}
	}
	*left = *right = FirstOffsetNumber; /* sentinel value, see dosplit() */

	v->spl_ldatum = PointerGetDatum(datum_l);
	v->spl_rdatum = PointerGetDatum(datum_r);

	return v;
}

/*
** Equality method
*/
bool *
g_bit_same(VarBit * b1, VarBit * b2, bool *result)
{
	if (bit_eq(b1, b2))
		*result = TRUE;
	else
		*result = FALSE;

	return (result);
}

/*
** SUPPORT ROUTINES
*/
bool
g_bit_leaf_consistent(VarBit * key,
					   VarBit * query,
					   StrategyNumber strategy)
{
	bool		retval;

	/*
	 * fprintf(stderr, "leaf_consistent, %d\n", strategy);
	 */
	switch (strategy)
	{
		case RTLeftStrategyNumber:
			retval = (bool) bit_left(key, query);
			break;
		case RTOverLeftStrategyNumber:
			retval = (bool) bit_over_left(key, query);
			break;
		case RTOverlapStrategyNumber:
			retval = (bool) bit_overlap(key, query);
			break;
		case RTOverRightStrategyNumber:
			retval = (bool) bit_over_right(key, query);
			break;
		case RTRightStrategyNumber:
			retval = (bool) bit_right(key, query);
			break;
		case RTSameStrategyNumber:
			retval = (bool) bit_eq(key, query);
			break;
		case RTContainsStrategyNumber:
			retval = (bool) bit_contains(key, query);
			break;
		case RTContainedByStrategyNumber:
			retval = (bool) bit_contained(key, query);
			break;
		default:
			retval = FALSE;
	}
	return (retval);
}

bool
g_bit_internal_consistent(VarBit * key,
						   VarBit * query,
						   StrategyNumber strategy)
{
	bool		retval;

	/*
	 * fprintf(stderr, "internal_consistent, %d\n", strategy);
	 */
	switch (strategy)
	{
		case RTLeftStrategyNumber:
		case RTOverLeftStrategyNumber:
			retval = (bool) bit_over_left(key, query);
			break;
		case RTOverlapStrategyNumber:
			retval = (bool) bit_overlap(key, query);
			break;
		case RTOverRightStrategyNumber:
		case RTRightStrategyNumber:
			retval = (bool) bit_right(key, query);
			break;
		case RTSameStrategyNumber:
		case RTContainsStrategyNumber:
			retval = (bool) bit_contains(key, query);
			break;
		case RTContainedByStrategyNumber:
			retval = (bool) bit_overlap(key, query);
			break;
		default:
			retval = FALSE;
	}
	return (retval);
}




	

bool
bit_eq(VarBit * arg1, VarBit * arg2)
{
	bool	   result;
	int			bitlen1,
				bitlen2,
				i;
	bits8	   *p1,
               *p2;
	
	bitlen1 = VARBITLEN(arg1);
	bitlen2 = VARBITLEN(arg2);

	// make sure arg1 is the larger of the two
	if (bitlen1 < bitlen2) {
	  VarBit *tmp = arg1;
	  int tmplen = bitlen1;
	  arg1 = arg2;
	  arg2 = tmp;
	  bitlen1 = bitlen2;
	  bitlen2 = tmplen;
	}

	p1 = VARBITS(arg1);
	p2 = VARBITS(arg2);
	result = TRUE;
	for (i = 0; i < VARBITBYTES(arg2); i++)
		result = result && (*p1++ == *p2++);
	for (; i < VARBITBYTES(arg1); i++)
		result = result && (*p1++ == 0);
	
	return(result);
}

VarBit *
bit_union(VarBit * arg1, VarBit * arg2)
{
	VarBit	   *result;
	int			len,
				bitlen1,
				bitlen2,
				i;
	bits8	   *p1,
			   *p2,
			   *r;
	bits8		mask;

	bitlen1 = VARBITLEN(arg1);
	bitlen2 = VARBITLEN(arg2);

	// make sure arg1 is the larger of the two
	if (bitlen1 < bitlen2) {
	  VarBit *tmp = arg1;
	  int tmplen = bitlen1;
	  arg1 = arg2;
	  arg2 = tmp;
	  bitlen1 = bitlen2;
	  bitlen2 = tmplen;
	}

	len = VARSIZE(arg1);
	result = (VarBit *) palloc(len);
	VARATT_SIZEP(result) = len;
	VARBITLEN(result) = bitlen1;

	p1 = VARBITS(arg1);
	p2 = VARBITS(arg2);
	r = VARBITS(result);
	for (i = 0; i < VARBITBYTES(arg2); i++)
		*r++ = *p1++ | *p2++;
	for (; i < VARBITBYTES(arg1); i++)
		*r++ = *p1++;
	
	/* Pad the result */
	mask = BITMASK << VARBITPAD(result);
	if (mask)
	{
		r--;
		*r &= mask;
	}

	return(result);
}

VarBit *
bit_inter(VarBit * arg1, VarBit * arg2)
{
	VarBit	   *result;
	int			len,
				bitlen1,
				bitlen2,
				i;
	bits8	   *p1,
			   *p2,
			   *r;
	bits8		mask;

	bitlen1 = VARBITLEN(arg1);
	bitlen2 = VARBITLEN(arg2);

	// make sure arg1 is the larger of the two
	if (bitlen1 < bitlen2) {
	  VarBit *tmp = arg1;
	  int tmplen = bitlen1;
	  arg1 = arg2;
	  arg2 = tmp;
	  bitlen1 = bitlen2;
	  bitlen2 = tmplen;
	}

	len = VARSIZE(arg2);
	result = (VarBit *) palloc(len);
	VARATT_SIZEP(result) = len;
	VARBITLEN(result) = bitlen2;

	p1 = VARBITS(arg1);
	p2 = VARBITS(arg2);
	r = VARBITS(result);
	for (i = 0; i < VARBITBYTES(arg2); i++)
		*r++ = *p1++ & *p2++;
	
	/* Pad the result */
	mask = BITMASK << VARBITPAD(result);
	if (mask)
	{
		r--;
		*r &= mask;
	}

	return(result);
}

VarBit *
bit_xor(VarBit * arg1, VarBit * arg2)
{
	VarBit	   *result;
	int			len,
				bitlen1,
				bitlen2,
				i;
	bits8	   *p1,
			   *p2,
			   *r;
	bits8		mask;

	bitlen1 = VARBITLEN(arg1);
	bitlen2 = VARBITLEN(arg2);

	// make sure arg1 is the larger of the two
	if (bitlen1 < bitlen2) {
	  VarBit *tmp = arg1;
	  int tmplen = bitlen1;
	  arg1 = arg2;
	  arg2 = tmp;
	  bitlen1 = bitlen2;
	  bitlen2 = tmplen;
	}

	len = VARSIZE(arg1);
	result = (VarBit *) palloc(len);
	VARATT_SIZEP(result) = len;
	VARBITLEN(result) = bitlen1;

	p1 = VARBITS(arg1);
	p2 = VARBITS(arg2);
	r = VARBITS(result);
	for (i = 0; i < VARBITBYTES(arg2); i++)
		*r++ = *p1++ ^ *p2++;
	for (; i < VARBITBYTES(arg1); i++)
		*r++ = *p1++;
	
	/* Pad the result */
	mask = BITMASK << VARBITPAD(result);
	if (mask)
	{
		r--;
		*r &= mask;
	}

	return(result);
}




int bit_count_bits(VarBit *arg) 
{
  bits8 *p = VARBITS(arg),
         mask;
  int i;
  
  int result = 0;

  for (i=0; i < VARBITBYTES(arg); i++,p++)
	for (mask=1;mask;mask<<=1)
	  result += (mask & *p ? 1 : 0);
  
  return result;
}


/* bit_size */
double *
bit_size(VarBit * a)
{
	double *result = palloc(sizeof(double));
	*result = (double)bit_count_bits(a);
	return result;
}

void
rt_bit_size(VarBit * a, double *size)
{
  *size = (double)bit_count_bits(a);
}




/* The following four methods compare the projections of the boxes
   onto the 0-th coordinate axis. These methods are useless for dimensions
   larger than 2, but it seems that R-tree requires all its strategies
   map to real functions that return something */

/*	is the right edge of (a) located to the left of
	the right edge of (b)? */
bool
bit_over_left(VarBit * a, VarBit * b)
{
	return TRUE;
}

/*	is the left edge of (a) located to the right of
	the left edge of (b)? */
bool
bit_over_right(VarBit * a, VarBit * b)
{
	return TRUE;
}


/* return 'true' if the projection of 'a' is
   entirely on the left of the projection of 'b' */
bool
bit_left(VarBit * a, VarBit * b)
{
	return FALSE;
}

/* return 'true' if the projection of 'a' is
   entirely on the right  of the projection of 'b' */
bool
bit_right(VarBit * a, VarBit * b)
{
	return FALSE;
}




/* Contains */
/* Box(A) CONTAINS Box(B) IFF pt(A) < pt(B) */
bool
bit_contains(VarBit * arg1, VarBit * arg2)
{
	bool	   result;
	int			len,
				bitlen1,
				bitlen2,
				i;
	bits8	   *p1,
               *p2;

	bitlen1 = VARBITLEN(arg1);
	bitlen2 = VARBITLEN(arg2);

	p1 = VARBITS(arg1);
	p2 = VARBITS(arg2);
	len = Min(VARBITBYTES(arg1),VARBITBYTES(arg2));

	result = TRUE;
	for (i = 0; i < len; i++)
		result = result && ( (~*p1++ & *p2++) == 0);
	for (; i < VARBITBYTES(arg2); i++)
		result = result && (*p2++ == 0);

	return(result);
}

bool
bit_contained(VarBit * a, VarBit * b)
{
	return bit_contains(b, a);
}

bool
bit_overlap(VarBit * a, VarBit * b)
{
	return bit_count_bits(bit_inter(a,b)) > 0;
}


double *
bit_distance(VarBit * a, VarBit * b)
{
	double *result = palloc(sizeof(double));
	*result = (double)bit_count_bits(bit_xor(a,b));
	return result;
}
